// This encapsulates the state we need to know how to configure a real
// serial port.

#include "SerialCfgState.h"
#include "Host.h"           	// to get config read/write routines
#include <string>

// ---------------------------------------------------------------------------
// public members
// ---------------------------------------------------------------------------

// default constructor
SerialCfgState::SerialCfgState()
{
    setDefaults();
}

// copy constructor
SerialCfgState::SerialCfgState(const SerialCfgState &obj)
{
    setTtyEnable( obj.getTtyEnable() );
    setPort( obj.getPort() );
    m_dcs = obj.m_dcs;
}


// assignment
SerialCfgState&
SerialCfgState::operator=(const SerialCfgState &rhs)
{
    // check for self-assignment
    if (this != &rhs) {
        setTtyEnable( rhs.getTtyEnable() );
        setPort( rhs.getPort() );
        m_dcs = rhs.m_dcs;
    }

    return *this;
}


// return copy of self
SerialCfgState*
SerialCfgState::clone() const
{
    return new SerialCfgState(*this);
}


// equality comparison
bool
SerialCfgState::operator==(const SerialCfgState &rhs) const
{
    return
        getTtyEnable() == rhs.getTtyEnable() &&
        getPort()      == rhs.getPort()      &&
        getBaudRate()  == rhs.getBaudRate()  &&
        getParity()    == rhs.getParity()    &&
        getBits()      == rhs.getBits()      &&
        getStopBits()  == rhs.getStopBits()  &&
        getFlowctl()   == rhs.getFlowctl()   &&
        // this shouldn't be necessary, but just in case
        m_dcs.stopbits == rhs.m_dcs.stopbits &&
        m_dcs.xonxoff  == rhs.m_dcs.xonxoff  ;
}

bool
SerialCfgState::operator!=(const SerialCfgState &rhs) const
{
    return !(*this == rhs);
}

void
SerialCfgState::setDefaults()
{
    m_tty_enabled = false;

    m_port = 1;  // com1

    m_dcs.baud = 110;
    m_dcs.wordlen = 8;
    m_dcs.parity = ctb::ParityNone;
    m_dcs.rtscts = false;

    // this is the one m_dcs member that we don't expose
    m_dcs.stopbits = 2;
    m_dcs.xonxoff = false;
}


// read from configuration file
void
SerialCfgState::Load(int term)
{
    char sg[10];
    sprintf(sg, "Term-%02d", term);
    std::string subgroup(sg);

    bool bv;
    int v;

    Host::ConfigReadBool(subgroup, "serial_port_enabled", &bv, false);
    setTtyEnable( bv );

    (void)Host::ConfigReadInt(subgroup, "com_port", &v, 1);
    v = (v < 1) ? 1 : (v > 4) ? 4 : v;
    setPort(v);

    (void)Host::ConfigReadInt(subgroup, "baud_rate", &v, 1);
    v = (v >= 2400) ? 2400
      : (v >= 1200) ? 1200
      : (v >=  300) ?  300
                    :  110;
    setBaudRate(v);

    (void)Host::ConfigReadInt(subgroup, "parity_mode", &v, 0);
    v = (v < 0) ? 0 : (v > 4) ? 4 : v;
    setParity(static_cast<ctb::Parity>(v));

    (void)Host::ConfigReadInt(subgroup, "byte_bits", &v, 8);
    v = (v < 7) ? 7 : (v > 8) ? 8 : v;
    setBits(v);

    (void)Host::ConfigReadInt(subgroup, "stop_bits", &v, 2);
    v = (v < 1) ? 1 : (v > 2) ? 2 : v;
    setStopBits(v);

    Host::ConfigReadBool(subgroup, "rtscts", &bv, false);
    setFlowctl( bv );
}


// save to configuration file
void
SerialCfgState::Save(int term) const
{
    char sg[10];
    sprintf(sg, "Term-%02d", term);
    std::string subgroup(sg);

    Host::ConfigWriteBool(subgroup, "serial_port_enabled", getTtyEnable() );
    Host::ConfigWriteInt( subgroup, "com_port",            getPort()      );
    Host::ConfigWriteInt( subgroup, "baud_rate",           getBaudRate()  );
    Host::ConfigWriteInt( subgroup, "parity_mode",         getParity()    );
    Host::ConfigWriteInt( subgroup, "byte_bits",           getBits()      );
    Host::ConfigWriteInt( subgroup, "stop_bits",           getStopBits()  );
    Host::ConfigWriteBool(subgroup, "rtscts",              getFlowctl()   );
}


void
SerialCfgState::setTtyEnable(bool enabled)
{
    m_tty_enabled = enabled;
}

bool
SerialCfgState::getTtyEnable() const
{
    return m_tty_enabled;
}


void
SerialCfgState::setPort(int com_port)
{
    m_port = com_port;
}

int
SerialCfgState::getPort() const
{
    return m_port;
}


void
SerialCfgState::setBaudRate(int baud_rate)
{
    m_dcs.baud = baud_rate;
}

int
SerialCfgState::getBaudRate() const
{
    return m_dcs.baud;
}


void
SerialCfgState::setParity(ctb::Parity mode)
{
    m_dcs.parity = mode;
}

ctb::Parity
SerialCfgState::getParity() const
{
    return m_dcs.parity;
}


void
SerialCfgState::setBits(int bits_per_byte)
{
    m_dcs.wordlen = bits_per_byte;
}

int
SerialCfgState::getBits() const
{
    return m_dcs.wordlen;
}


void
SerialCfgState::setStopBits(int stop_bits)
{
    assert(stop_bits == 1 || stop_bits == 2);
    m_dcs.stopbits = stop_bits;
}

int
SerialCfgState::getStopBits() const
{
    return m_dcs.stopbits;
}


void
SerialCfgState::setFlowctl(bool rts_cts_enabled)
{
    m_dcs.rtscts = rts_cts_enabled;
}

bool
SerialCfgState::getFlowctl() const
{
    return m_dcs.rtscts;
}


ctb::SerialPort_DCS *
SerialCfgState::getDCS()
{
    return &m_dcs;
}
