// =============================================================
// system configuration class
// =============================================================

#include "Host.h"
#include "SystemCfg.h"
#include "SerialCfgState.h"

// default constructor
SystemCfg::SystemCfg()
{
    Load(); // initialize the state based on the config file
}


// assignment
SystemCfg& SystemCfg::operator=(const SystemCfg &rhs)
{
    m_cpu_speed_regulated = rhs.m_cpu_speed_regulated;
    m_ram_kb              = rhs.m_ram_kb;
    m_num_terminals       = rhs.m_num_terminals;
    m_num_cassette_drives = rhs.m_num_cassette_drives;

    for(unsigned int n=0; n<m_serial.size(); n++) {
        *m_serial[n] = *rhs.m_serial[n];
    }

    return *this;
}


// copy constructor -- performs deep copy
SystemCfg::SystemCfg(const SystemCfg &obj)
{
    //*this = obj;
    SetCpuSpeedRegulated( obj.GetCpuSpeedRegulated() );
    SetRamKB(             obj.GetRamKB()             );
    SetNumTerminals(      obj.GetNumTerminals()      );
    SetNumCassetteDrives( obj.GetNumCassetteDrives() );

    m_serial.clear();  // delete anything we have
    for(unsigned int n=0; n<obj.m_serial.size(); n++) {
        m_serial.push_back(obj.GetSerialCfgState(n)->clone());
    }
}


SystemCfg::~SystemCfg()
{
    // nothing to do
}


// equality comparison
bool
SystemCfg::operator==(const SystemCfg &rhs) const
{
    if ((GetCpuSpeedRegulated() != rhs.GetCpuSpeedRegulated())  ||
        (GetRamKB()             != rhs.GetRamKB())              ||
        (GetNumTerminals()      != rhs.GetNumTerminals())       ||
        (GetNumCassetteDrives() != rhs.GetNumCassetteDrives()))
        return false;

    // check terminals
    for(unsigned int n=0; n<m_serial.size(); n++) {
        if (*GetSerialCfgState(n) != *rhs.GetSerialCfgState(n)) {
            return false;
        }
    }

    return true;
}

bool
SystemCfg::operator!=(const SystemCfg &rhs) const
{
    return !(*this == rhs);
}


void
SystemCfg::SetCpuSpeedRegulated(bool v)
{
    m_cpu_speed_regulated = v;
}


void
SystemCfg::SetRamKB(int v)
{
    assert(v >= 4 && v <= 64);
    m_ram_kb = v;
}


void
SystemCfg::SetNumTerminals(int v)
{
    assert(v >= 1 && v <= 16);
    m_num_terminals = v;
}


void
SystemCfg::SetNumCassetteDrives(int v)
{
    assert(v >= 0 && v <= 16);
    assert((v & 1) == 0);
    m_num_cassette_drives = v;
}

void
SystemCfg::SetSerialCfg(int n, SerialCfgState &cfg)
{
    if (m_serial[n] != NULL) {
        free(m_serial[n]);
        m_serial[n] = NULL;
    }
    m_serial[n] = cfg.clone();
}

// read from configuration file
void
SystemCfg::Load()
{
    std::string subgroup("..");

    bool bv;
    int v;

    Host::ConfigReadBool(subgroup, "cpu_speed_regulated", &bv, true);
    SetCpuSpeedRegulated( bv );

    (void)Host::ConfigReadInt(subgroup, "ram_kb", &v, 64);
    v = (v < 4) ? 4 : (v > 64) ? 64 : v;
    SetRamKB( v );

    (void)Host::ConfigReadInt(subgroup, "terminals", &v, 1);
    v = (v < 1) ? 1 : (v > 16) ? 16 : v;
    SetNumTerminals( v );

    (void)Host::ConfigReadInt(subgroup, "cassettes", &v, 2);
    v = (v + 1) & ~1;   // round up to next even (drives are in pairs)
    v = (v < 0) ? 0 : (v > 16) ? 16 : v;
    SetNumCassetteDrives( v );

    // load terminal serial port configuration
    assert(m_serial.empty()); // we expect Load() to be called once
    for(int n=0; n<MAX_TERMINALS; n++) {
        m_serial.push_back(new SerialCfgState());
        m_serial[n]->Load(n);
    }
}


// save to configuration file
void
SystemCfg::Save() const
{
    std::string subgroup("..");

    Host::ConfigWriteBool(subgroup, "cpu_speed_regulated", GetCpuSpeedRegulated() );
    Host::ConfigWriteInt( subgroup, "ram_kb",    GetRamKB() );
    Host::ConfigWriteInt( subgroup, "terminals", GetNumTerminals() );
    Host::ConfigWriteInt( subgroup, "cassettes", GetNumCassetteDrives() );

    // save terminal serial port configuration
    for(unsigned int n=0; n<m_serial.size(); n++) {
        m_serial[n]->Save(n);
    }
}

