// this code emulate a Wang 3320 terminal controller connected to a
// Wang 3315 terminal, which is a rebadged ASR33 TTY.
//
// Mostly this just passes requests through to the appropriate subunit.

#include "system3300.h"
#include "Term3315.h"
#include "Term3315_impl.h"
#include "SerialCfgState.h"
#include "UiTerm3315.h"
#include "UiSystem.h"

#include <sstream>


// create a new terminal emulation core
Term3315::Term3315(Cpu3300 *cpu, Scheduler &scheduler, int termnum) :
    m_ui(0),
    m_kbd(0),
    m_prt(0),
    m_pun(0),
    m_serport(NULL),    // by default no associated com port
    m_scs(NULL),
    m_initialized(false)
{
    assert(termnum >= 0 && termnum < 16);

    const int station = (termnum >> 2); // by convention

    // create components
    m_ui  = new UiTerm3315(termnum, this);
    m_pun = new Term3315_pun(termnum, m_ui);
    m_kbd = new Term3315_kbd(cpu, scheduler, m_pun, termnum, station);
    m_prt = new Term3315_prt(cpu, scheduler, m_pun, termnum, station, m_ui, this);

    // tell m_pun which keyboard it is associated with
    m_pun->BindKeyboard(m_kbd);

    m_scs = new SerialCfgState();
    m_scs->setTtyEnable(false);

    m_initialized = true;
}


Term3315::~Term3315()
{
    delete m_prt;
    delete m_kbd;
    delete m_pun;
    if (m_serport != NULL) {
        m_serport->Close();
        delete m_serport;
        m_serport = NULL;
    }
    delete m_scs;

    m_ui->Destroy();    // it is a wxWindow, it must be Destroyed
}


void
Term3315::Reset()
{
    assert(Initialized());
    m_kbd->Reset();
    m_prt->Reset();
    m_pun->Reset();
}

void
Term3315::Clear()
{
    assert(Initialized());
    m_kbd->Clear();
    m_prt->Clear();
    m_pun->Clear();
}

void
Term3315::WrControl(int addr, int data)
{
    assert(Initialized());
    IoDevice *subunit = (addr & 1) ? static_cast<IoDevice*>(m_prt)
                                   : static_cast<IoDevice*>(m_kbd);
    subunit->WrControl(addr, data);
}

int
Term3315::RdStatus(int addr) const
{
    assert(Initialized());
    IoDevice *subunit = (addr & 1) ? static_cast<IoDevice*>(m_prt)
                                   : static_cast<IoDevice*>(m_kbd);
    return subunit->RdStatus(addr);
}

void
Term3315::WrData(int addr, int data)
{
    assert(Initialized());
    IoDevice *subunit = (addr & 1) ? static_cast<IoDevice*>(m_prt)
                                   : static_cast<IoDevice*>(m_kbd);
    subunit->WrData(addr, data);
}

int
Term3315::RdData(int addr) const
{
    assert(Initialized());
    IoDevice *subunit = (addr & 1) ? static_cast<IoDevice*>(m_prt)
                                   : static_cast<IoDevice*>(m_kbd);
    return subunit->RdData(addr);
}

void
Term3315::IAck(int addr)
{
    assert(Initialized());
    IoDevice *subunit = (addr & 1) ? static_cast<IoDevice*>(m_prt)
                                   : static_cast<IoDevice*>(m_kbd);
    subunit->IAck(addr);
}

// ======================================================================
// functions not part of IoDevice inheritance
// ======================================================================

// mechanism for UI to notify us of keystroke to our terminal
void
Term3315::Keystroke(uint8 keycode)
{
    if (!Initialized())
        return;
    m_kbd->Keystroke(keycode, Term3315_kbd::src_emu_tty);
}

// mechanism to allow the term to poll the serial port
void
Term3315::PollCheck()
{
    if (m_serport == NULL)
        return;

    // because emulated time is bursty and doesn't track real time very well
    // on the short scale, and because the serial port i/o stream may be
    // buffered, we can't count on the delay between receiving two characters
    // on a serial port to be slower than the rate the emulator is draining
    // them.  thus, don't drain the serial port fifo if the emulator thinks
    // it is receiving the previous character.
    if (m_kbd->receiving())
        return;

    int cnt = 0;
    int stat = m_serport->Ioctl(ctb::CTB_SER_GETINQUE, &cnt);
    if (stat == 0 && cnt > 0) {
        char ch;
        cnt = m_serport->Read(&ch, 1);
        if (cnt == 1) {
extern FILE *g_outf;
if (g_outf) { fprintf(g_outf, "poll: got 0x%x ('%c')\n", (int)ch, ch); fflush(g_outf); }
            m_kbd->Keystroke((uint8)ch, Term3315_kbd::src_serial);
        }
    }
}


// ======================================================================
// real tty interface
// ======================================================================

// return false on failure
bool
Term3315::setTtyConfiguration(SerialCfgState *cfg)
{
    *m_scs = *cfg;       // stash a copy

    if (m_serport != NULL) {
        m_serport->Close();
        delete m_serport;
        m_serport = NULL;
        m_prt->SetTty(NULL); // remove serial port association from printer
    }

    if (! cfg->getTtyEnable()) {
        return true; // nothing to make
    }

    m_serport = new ctb::SerialPort();
    if (m_serport == NULL) {
        UiError("Couldn't create a ctb::SerialPort()");
        return false;
    }

    std::ostringstream pn;
    pn << "com" << cfg->getPort();
    std::string portname(pn.str());

    int sps = m_serport->Open(portname.c_str(), cfg->getDCS());
    if (sps < 0) {
        UiError("Couldn't open COM%d, status=%d", cfg->getPort(), sps);
        free(m_serport);
        m_serport = NULL;
        return false;
    }

    // tell printer which serial port to echo to
    m_prt->SetTty(m_serport);

    return true;
}

SerialCfgState *
Term3315::getTtyConfiguration() const
{
    return m_scs;
}

// ======================================================================
// proxy down to the pun subunit
// ======================================================================

void
Term3315::ReaderStart()
{
    assert(Initialized());
    m_pun->ReaderStart();
}

void
Term3315::ReaderStop()
{
    assert(Initialized());
    m_pun->ReaderStop();
}

bool
Term3315::ReaderEnabled() const
{
    if (!Initialized()) return false;
    return m_pun->ReaderEnabled();
}

void
Term3315::SetPunchEnabled(bool enable)
{
    assert(Initialized());
    m_pun->SetPunchEnabled(enable);
}

bool
Term3315::PunchEnabled() const
{
    if (!Initialized()) return false;
    return m_pun->PunchEnabled();
}

// indiate if a tape is already mounted or not
bool
Term3315::TapeMounted(const unit_t unit) const
{
    if (!Initialized()) return false;
    return m_pun->TapeMounted(unit);
}


// mount a tape for subsequent reading/writing
void
Term3315::MountTape(const unit_t unit, const char *filename)
{
    assert(Initialized());
    m_pun->MountTape(unit, filename);
}

// unmount the currently mounted tape
void
Term3315::UnmountTape(const unit_t unit)
{
    assert(Initialized());
    m_pun->UnmountTape(unit);
}


char *
Term3315::GetTapePath(const unit_t unit) const
{
    if (!Initialized()) return "";
    return m_pun->GetTapePath(unit);
}

// return position of r/w head relative to start, in bytes
int
Term3315::TapePosition(const unit_t unit)
{
    if (!Initialized()) return 0;
    return m_pun->TapePosition(unit);
}


// move r/w position to given point on the tape, in bytes
void
Term3315::SetTapePosition(const unit_t unit, int position)
{
    assert(Initialized());
    m_pun->SetTapePosition(unit, position);
}
