/////////////////////////////////////////////////////////////////////////////
// UiSystem.cpp
// Modified from wxWindows sample program, drawing.cpp, by Robert Roebling.
//
// This file contains the entry point for the emulator.
/////////////////////////////////////////////////////////////////////////////

// ============================================================================
// declarations
// ============================================================================

// ----------------------------------------------------------------------------
// headers
// ----------------------------------------------------------------------------

#include "UiSystem.h"
#include "Host.h"
#include "UiSystemCfgDlg.h"
#include "UiFrontPanel.h"
#include "UiMyAboutDlg.h"
#include "SystemCfg.h"
#include "SerialCfgState.h"

#include "system3300.h"
#include "Term3315.h"

#include "wx/cmdline.h"         // req'd by wxCmdLineParser
#include "wx/tokenzr.h"         // req'd by wxStringTokenizer
#include "wx/settings.h"        // req'd by GetMetric()

// ============================================================================
// implementation
// ============================================================================

// Create a new application object: this macro will allow wxWindows to
// create the application object during program execution (it's better
// than using static object for many reasons) and also declares the
// accessor function wxGetApp() which will return the reference of the
// right type (i.e. MyApp and not wxApp)
IMPLEMENT_APP(MyApp)

BEGIN_EVENT_TABLE(MyApp, wxApp)
    EVT_IDLE (MyApp::OnIdle)
END_EVENT_TABLE()

// ----------------------------------------------------------------------------
// the application class
// ----------------------------------------------------------------------------

wxFileConfig* MyApp::m_config;          // file configuration object

MyApp::term_state_t MyApp::m_term_state = MyApp::RUNNING;
bool MyApp::m_do_reconfig = false;

// `Main program' equivalent: the program execution "starts" here
bool
MyApp::OnInit()
{
#if 0
    // read wxToolBar notes to see why this is done
    if (wxTheApp->GetComCtl32Version() >= 600 && ::wxDisplayDepth() >= 32)
        wxSystemOptions::SetOption(wxT("msw.remap"), 2);
#endif
    SetTermState(RUNNING);

    // configuration file stuff
    SetVendorName("Slash & Burn Software");     // not actually used by wxFileConfig
    SetAppName("emu3300");

    // set up emulator state
    System3300 sys;
    sys.Init();

    // create the frame containing the front panel interface
    SetTopWindow(new UiFrontPanel());

    // must call base class version to get command line processing
    // if FALSE, the app terminates
    return wxApp::OnInit();
}


// the user requests a change in configuration from the UiFrontPanel.
// however, doing so often requires a tear down and rebuild of all the
// components.  destroying the frontpanel instance and then returning
// to it is uncouth.  instead, when the user wants to reconfigure, this
// function is called and that desire is noted, and we return immediately.
// later on, in the OnIdle code, this flag is checked and the reconfiguration
// happens then.
void
MyApp::Reconfigure()
{
    m_do_reconfig = true;
}


// this is called whenever the wxWidget event queue is empty,
// indicating there is no work pending.
void
MyApp::OnIdle(wxIdleEvent &event)
{
    System3300 sys;

    if (m_do_reconfig) {
        m_do_reconfig = false;
        SystemCfg sys_cfg(*sys.GetConfig());
        SystemCfgDlg dlg(NULL, &sys_cfg);
        int status = dlg.ShowModal();
        if (status == SystemCfgDlg::CHANGED) {
            // tear down the old config and build the new one
            System3300::SetConfig(sys_cfg);
        }
    }

    switch (GetTermState()) {
        case RUNNING:
            // this is the normal case during emulation
            sys.EmulateTimeslice();
            event.RequestMore(true);        // give more idle events
            break;
        case TERMINATING:
            // we've been signalled to shut down the universe.
            // change the flag to know we've already cleaned up, in case
            // we receive another OnIdle call.
            SetTermState(TERMINATED);
            sys.Cleanup();
            (UiFrontPanel*)(wxTheApp->GetTopWindow())->Destroy();
            break;
        case TERMINATED:
            // do nothing -- we already requested a cleanup
            break;
        default:
            wxFAIL;
            break;
    }
}


// called to signal we want to end things.  at this point nothing has been
// shut down.  This can be called from within a subordinate window; we
// can't destroy anything now, otherwise we'd return to an object that was
// deleted already.  Instead, we just set a flag indicating we want to shut
// down, and the OnIdle() handler will clean things up from a safe point later.
void
MyApp::Terminate()
{
    // flag the OnIdle handler to clean up next time it gets a chance
    SetTermState(TERMINATING);
}


// This is called just before quitting the entire app,
// but before wxWindows cleans up its internal resources.
int
MyApp::OnExit()
{
    // clean up the singleton state
    Host().terminate();

    return 0;
}


// set the command line parsing options
void
MyApp::OnInitCmdLine(wxCmdLineParser& parser)
{
    // set default wx options
    wxApp::OnInitCmdLine(parser);

    parser.DisableLongOptions();        // -foo, not --foo

    // add options specific to this app
    parser.AddOption("s", "script", "script file to load on startup", wxCMD_LINE_VAL_STRING);
}


// after the command line has been parsed, decode what it finds
bool
MyApp::OnCmdLineParsed(wxCmdLineParser& parser)
{
    // let base class handle its defaults
    bool OK = wxApp::OnCmdLineParsed(parser);

    if (OK) {
        // do my processing here
        // UiInfo("Command line items: %d", parser.GetParamCount());
#ifdef WANG2200
        wxString filename;
        if (parser.Found("s", &filename)) {
            const int io_addr = 0x01;   // default keyboard device
            IOcard_keyboard::invoke_script(io_addr, filename.c_str());
        }
#endif
    }

    return OK;
}


// ========================================================================
// Help/About functions
// ========================================================================

wxMenu*
MyApp::MakeHelpMenu()
{
    wxMenu *menuHelp = new wxMenu;
//  menuHelp->Append(Help_Quickstart,  "&Quickstart",      "Information for new users of WangEmu");
//  menuHelp->Append(Help_Configure,   "&Configuration",   "Information about configuring the emulator");
//  menuHelp->Append(Help_Keyboard,    "&Keyboard",        "Information about how the 2200 keyboard is mapped onto yours");
//  menuHelp->Append(Help_Menus,       "&Menus",           "Information about the emulator menu system");
//  menuHelp->Append(Help_Printer,     "&Printer",         "Information about the using the emulated printer");
//  menuHelp->Append(Help_Script,      "&Script",          "Information about loading files into the emulator");
    menuHelp->Append(Help_Website,     "&Website",         "Open a browser to the emulator's web site");
    #ifndef __WXMAC__
    // this isn't needed on the Mac because the About menu item gets forcibly moved to another place
    menuHelp->AppendSeparator();
    #endif
    menuHelp->Append(Help_About,       "&About...",        "Information about the program");

    return menuHelp;
}


void
MyApp::OnHelp_Launcher(wxCommandEvent& event)
{
    wxString helpfile;
    bool absolute = FALSE;
    switch (event.GetId()) {
//      case Help_Quickstart:   helpfile = "quickstart.html";      break;
//      case Help_Configure:    helpfile = "configure.html";       break;
//      case Help_Keyboard:     helpfile = "keyboard.html";        break;
//      case Help_Menus:        helpfile = "menus.html";           break;
//      case Help_Printer:      helpfile = "printer.html";         break;
//      case Help_Script:       helpfile = "script.html";          break;
        case Help_Website:      helpfile = "http://www.wang3300.org/"; absolute = true; break;
        default:
            wxFAIL;
            return;
    }

    wxChar sep(wxFileName::GetPathSeparator());
    wxString target_file = (absolute) ? helpfile
                                      : ( "file://" + Host::getAppHome() + sep + "html" + sep + helpfile );
    ::wxLaunchDefaultBrowser(target_file);
}


void
MyApp::OnHelp_About(wxCommandEvent& WXUNUSED(event))
{
#if 0
    wxString msg;
    msg.Printf( "Wang 3300 Emulator\n"
                "Version 0.4d, January 31, 2012\n"
                "Copyright (c) Jim Battle, Slash & Burn Software, 2012\n\n"
                "Built with %s\n\n"
                "Thanks for crucial information and encouragement:\n"
                "    Steve Witham\n"
                "    Tom Lake"
                , wxVERSION_STRING
              );

    // unfortunately, centering the text doesn't work on windows
    wxMessageDialog dialog(NULL, msg, "About this program",
                            wxOK | wxICON_INFORMATION | wxCENTRE);
#ifndef __WXMSW__
    dialog.CenterOnParent();    // this doesn't seem to work (on Win32 at least)
#endif
    dialog.ShowModal();
#else
    MyAboutDlg dlg(NULL);
    dlg.ShowModal();
#endif
}

// ============================================================================
// alert messages
// ============================================================================

// shared code
static bool
UI_AlertMsg(long style, wxString title, const char *fmt, va_list &args)
{
    char buff[1000];
    vsnprintf(buff, sizeof(buff), fmt, args);

    wxString info(buff);
    wxMessageDialog dialog(NULL, info, title, style);
    int rv = dialog.ShowModal();
    return (rv == wxID_YES);
}

// error icon
void
UiError(const char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    (void)UI_AlertMsg( wxOK | wxICON_ERROR, "Error", fmt, args );
    va_end(args);
}

// exclamation icon
void
UiWarn(const char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    (void)UI_AlertMsg( wxOK | wxICON_EXCLAMATION, "Warning", fmt, args );
    va_end(args);
}

void
UiInfo(const char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    (void)UI_AlertMsg( wxOK | wxICON_INFORMATION, "Information", fmt, args );
    va_end(args);
}


// get a YES/NO confirmation.  return true for yes.
bool
UiConfirm(const char *fmt, ...)
{
    long style = wxYES | wxNO | wxNO_DEFAULT | wxICON_EXCLAMATION;
//  long style = wxYES | wxNO | wxNO_DEFAULT | wxICON_QUESTION;

    va_list args;
    va_start(args, fmt);
    bool rv = UI_AlertMsg( style, "Question", fmt, args );
    va_end(args);

    return rv;
}

// ========================================================================
//    User Interface utility functions
// ========================================================================

// called periodically by the core emulation so UI events can occur
void
UiYield(void)
{
    wxTheApp->Yield(true);      // simply returns if things get recursive
}


// set simulation time for informative display
void
UiSetSimSeconds(unsigned long secs, float relative_speed)
{
    wxString str;
    wxString format = (relative_speed >= 10.0f) ?
                      "Sim time: %d seconds, %3.0fx"
                    : "Sim time: %d seconds, %3.1fx";
    str.Printf( format, secs, relative_speed );

    UiFrontPanel *tf = (UiFrontPanel*)(wxTheApp->GetTopWindow());
    tf->m_statbar->SetStatusText(str);
}
