// ----------------------------------------------------------------------------
// headers
// ----------------------------------------------------------------------------

#include "system3300.h"
#include "UiSystem.h"
#include "Host.h"
#include "UiSystemCfgDlg.h"
#include "UiSerialCfgDlg.h"
#include "SystemCfg.h"

enum
{
    ID_MEMSIZE_CHOICE = 1,
    ID_SPEED_REGULATED,
    ID_NUM_TERMINALS,
    ID_NUM_CASSETTE_DRIVES,
    ID_SERIAL_TERM_0,
    // ...
    ID_SERIAL_TERM_MAX = ID_SERIAL_TERM_0+MAX_TERMINALS-1,

    ID_BTN_REVERT = 300
};


// dialog events to catch
BEGIN_EVENT_TABLE(SystemCfgDlg, wxDialog)
    EVT_CHOICE(ID_MEMSIZE_CHOICE,       SystemCfgDlg::OnMemsizeChoice)
    EVT_CHECKBOX(ID_SPEED_REGULATED,    SystemCfgDlg::OnSpeedRegulated)
    EVT_CHOICE(ID_NUM_TERMINALS,        SystemCfgDlg::OnNumTerminals)
    EVT_CHOICE(ID_NUM_CASSETTE_DRIVES,  SystemCfgDlg::OnNumCassetteDrives)
    EVT_COMMAND_RANGE(ID_SERIAL_TERM_0, ID_SERIAL_TERM_MAX,
        wxEVT_COMMAND_BUTTON_CLICKED,   SystemCfgDlg::OnSerialCfg)
    EVT_BUTTON(wxID_ANY,                SystemCfgDlg::OnButton)
END_EVENT_TABLE()


SystemCfgDlg::SystemCfgDlg(wxFrame *parent, SystemCfg *cfg) :
        wxDialog(parent, -1, wxT("System Configuration"),
                 wxDefaultPosition, wxDefaultSize,
                 wxDEFAULT_DIALOG_STYLE | wxRESIZE_BORDER),
        m_oldcfg(cfg)
{
    const int h_text_margin = 8;

    // make a copy of the config state passed to us
    m_config = new SystemCfg(*cfg);

    hSpeedRegulated = new wxCheckBox(this, ID_SPEED_REGULATED,
                                           wxT("Regulated CPU Speed"));

    {
        hMemsize = new wxChoice(this, ID_MEMSIZE_CHOICE);
        static const int typical_ram_sizes[] =
            { 4, 8, 12, 16, 20, 24, 28, 32, 40, 48, 56, 64 };
        for(int n=0; n < sizeof(typical_ram_sizes)/sizeof(int); n++) {
            wxString label;
            label.Printf(wxT("%2d KB"), typical_ram_sizes[n]);
            hMemsize->Append(label, (void*)typical_ram_sizes[n]);
        }
    }

    {
        hNumTerms = new wxChoice(this, ID_NUM_TERMINALS);
        for(int n=1; n <= MAX_TERMINALS; n++) {
            wxString label;
            label.Printf(wxT("%1d"), n);
            hNumTerms->Append(label, (void*) n);
        }
    }

    {
        hNumCassettes = new wxChoice(this, ID_NUM_CASSETTE_DRIVES);
        for(int n=0; n <= 8; n += 2) {
            wxString label;
            label.Printf(wxT("%1d"), n);
            hNumCassettes->Append(label, (void*) n);
        }
    }

    wxFlexGridSizer *grid = new wxFlexGridSizer(4, 2, 2, 0);

    grid->Add(new wxStaticText(this, -1, wxT("RAM")), 0,
                wxALIGN_RIGHT | wxALIGN_CENTER_VERTICAL | wxRIGHT, h_text_margin);
    grid->Add(hMemsize, 0);

    grid->Add(new wxStaticText(this, -1, wxT("Terminals")), 0,
                wxALIGN_RIGHT | wxALIGN_CENTER_VERTICAL | wxRIGHT, h_text_margin);
    grid->Add(hNumTerms, 0);

    wxBoxSizer *termbutton_sizer = new wxBoxSizer(wxVERTICAL);
    for(int n=0; n < MAX_TERMINALS; n++) {
        wxString label;
        label.Printf("Config Term #%d", n+1);
        hSerialCfg[n] = new wxButton(this, ID_SERIAL_TERM_0 + n, label);
        termbutton_sizer->Add(hSerialCfg[n], 0, wxALL, 0);
    }
    grid->Add(new wxStaticText(this, -1, wxT("")), 0,
                wxALIGN_RIGHT | wxALIGN_CENTER_VERTICAL | wxRIGHT, h_text_margin);
    grid->Add(termbutton_sizer, 0);

    grid->Add(new wxStaticText(this, -1, wxT("Cassette Drives")), 0,
                wxALIGN_RIGHT | wxALIGN_CENTER_VERTICAL | wxRIGHT, h_text_margin);
    grid->Add(hNumCassettes, 0);

    wxBoxSizer *ctlsizer = new wxBoxSizer(wxVERTICAL);
    ctlsizer->Add(hSpeedRegulated, 0, wxALL, 5);
    ctlsizer->Add(grid, 0, wxALL, 5);

#if 0
    // and this option
    hWarnIO = new wxCheckBox(this, ID_CHK_WARN_IO,
                                   wxT("Warn on Invalid IO Device Access"));
#endif

    // put two buttons side by side
    hBtnRevert = new wxButton(this, ID_BTN_REVERT, wxT("Revert"));
    hBtnOK     = new wxButton(this, wxID_OK,       wxT("OK"));
    hBtnCancel = new wxButton(this, wxID_CANCEL,   wxT("Cancel"));

    wxBoxSizer *button_sizer = new wxBoxSizer( wxHORIZONTAL );
    button_sizer->Add(hBtnRevert, 0, wxALL, 10);
    button_sizer->Add(hBtnOK,     0, wxALL, 10);
    button_sizer->Add(hBtnCancel, 0, wxALL, 10);
#ifdef __WXMAC__
    // the cancel button was running into the window resizing grip
    button_sizer->AddSpacer(10);
#endif
    hBtnRevert->Disable();      // until something changes

    // config grids on top, confirmation buttons on the bottom
    wxBoxSizer *topsizer = new wxBoxSizer( wxVERTICAL );
    topsizer->Add(ctlsizer,     1, wxEXPAND);           // vertically stretchable
    topsizer->Add(button_sizer, 0, wxALIGN_RIGHT);      // vertically unstretchable

    UpdateDlg();        // select current options

    // tell the thing to get to work
    SetSizer(topsizer);                 // use the sizer for layout
    topsizer->SetSizeHints(this);       // set size hints to honor minimum size

    GetDefaults();              // get default size & location
}


// update the display to reflect the current state
void
SystemCfgDlg::UpdateDlg()
{
    int regulated = m_config->GetCpuSpeedRegulated();
    hSpeedRegulated->SetValue(regulated ? 1 : 0);

    int ram_kb = m_config->GetRamKB();
    int n = (ram_kb <= 32) ? (ram_kb/4 - 1) : (ram_kb/8 + 3);
    hMemsize->SetSelection(n);

    int terms = m_config->GetNumTerminals();
    hNumTerms->SetSelection(terms - 1);

    int cassettes = m_config->GetNumCassetteDrives();
    hNumCassettes->SetSelection(cassettes/2);

    // show configuration buttons only for the available terminals
    for(int n=0; n<MAX_TERMINALS; n++) {
        hSerialCfg[n]->Show(n < terms);
    }

#if 0
    int warnio = iomgr_get_warnio();
    hWarnIO->SetValue(warnio == 1);
#endif

    bool changed = (*m_oldcfg != *m_config);
    hBtnRevert->Enable(changed);
}


void
SystemCfgDlg::OnSpeedRegulated( wxCommandEvent& WXUNUSED(event) )
{
    bool checked = hSpeedRegulated->IsChecked();
    m_config->SetCpuSpeedRegulated(checked);
    UpdateDlg();
}


void
SystemCfgDlg::OnMemsizeChoice( wxCommandEvent& WXUNUSED(event) )
{
    int selection = hMemsize->GetSelection();
    m_config->SetRamKB(int(hMemsize->GetClientData(selection)));
    UpdateDlg();
}


void
SystemCfgDlg::OnNumTerminals( wxCommandEvent& WXUNUSED(event) )
{
    int selection = hNumTerms->GetSelection();
    m_config->SetNumTerminals(int(hNumTerms->GetClientData(selection)));
    UpdateDlg();
    RefitDlg();
}

// changing the number of terminals (either explicitly, or via Revert,
// may require resizing the dialog, as the number of "Config Tern #n"
// buttons may change.
void
SystemCfgDlg::RefitDlg()
{
    Layout();
    Fit();
    wxSize sz(GetSize());
    SetMinSize(sz);
}

void
SystemCfgDlg::OnNumCassetteDrives( wxCommandEvent& WXUNUSED(event) )
{
    int selection = hNumCassettes->GetSelection();
    m_config->SetNumCassetteDrives(int(hNumCassettes->GetClientData(selection)));
    UpdateDlg();
}


#if 0
void
SystemCfgDlg::OnWarnIO( wxCommandEvent& WXUNUSED(event) )
{
    bool checked = hWarnIO->IsChecked();
    iomgr_set_warnio(checked ? 1 : 0);
    UpdateDlg();
}
#endif


// used for all dialog button presses
void
SystemCfgDlg::OnSerialCfg(wxCommandEvent& event)
{
    int idx = event.GetId() - ID_SERIAL_TERM_0;
    SerialCfgState *cfg = m_config->GetSerialCfgState(idx);

    SerialCfgDlg *hSerial = new SerialCfgDlg(this, cfg);
    int status = hSerial->ShowModal();

    UpdateDlg();
}

// used for all dialog button presses
void
SystemCfgDlg::OnButton(wxCommandEvent& event)
{
    bool changed = (*m_oldcfg != *m_config);

    switch (event.GetId()) {

        case ID_BTN_REVERT:
            *m_config = *m_oldcfg;   // revert to original state
            UpdateDlg();
            RefitDlg();
            hBtnRevert->Disable();
            break;

        case wxID_OK:
            SaveDefaults();         // save location & size of dlg
            if (changed) {
                *m_oldcfg = *m_config;  // send working copy back to real state
            }
            delete m_config; m_config = NULL;
            EndModal((changed) ? CHANGED : UNCHANGED);
            break;

        case wxID_CANCEL:
            SaveDefaults();         // save location & size of dlg
            delete m_config; m_config = NULL;
            EndModal(CANCELLED);
            break;

        default:
            assert(0);
            break;
    }
}


// save dialog geometry to the config file
void
SystemCfgDlg::SaveDefaults()
{
    std::string subgroup("ui/configdlg");

    // save position and size
    Host::ConfigWriteWinGeom(this, subgroup);
}


void
SystemCfgDlg::GetDefaults()
{
    // see if we've established a favored location and size
    std::string subgroup("ui/configdlg");
    Host::ConfigReadWinGeom(this, subgroup);
}
